/*
 * Part of the PiyoPiyoJ project.
 * (PiyoPiyo Java)
 *
 *        DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *                    Version 2, December 2004 
 *
 * Modified by gamemanj <asdd2808@gmail.com> to remove expletives
 * Original copyright (C) 2004 Sam Hocevar <sam@hocevar.net> 
 *
 * Everyone is permitted to copy and distribute verbatim or modified 
 * copies of this license document, and changing it is allowed as long 
 * as the name is changed. 
 *
 *            DO WHATEVER YOU WANT TO PUBLIC LICENSE 
 *   TERMS AND CONDITIONS FOR COPYING, DISTRIBUTION AND MODIFICATION 
 *
 *  0. You just DO WHATEVER YOU WANT TO.
 */
package piyopiyoj;

import gabien.GaBIEn;
import gabien.IRawAudioDriver;
import gabien.SimpleMixer;
import gabien.music.PiyoPiyoFile;
import gabien.music.PiyoPiyoPlayer;
import gabien.music.PiyoPiyoPlayingSample;
import gabien.ui.*;
import piyopiyoj.games.NotesheetGame;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.LinkedList;
import java.util.concurrent.locks.ReentrantLock;

/**
 * Used to be ripped straight out of IME,
 * but the 'Game' pattern really doesn't work that well when mixing in gabien.ui components.
 * If I port IME, I'll probably make a way to 'host' gabien.ui components without going all-in,
 * but PiyoPiyoJ is too UI-heavy for that to really be viable
 */
public class Game {
    private UIGame outerGame;
    public final SimpleMixer coreMixer = new SimpleMixer(true, true);
    public final PiyoPiyoFile coreFile = new PiyoPiyoFile();
    public final PiyoPiyoPlayer corePlayer = new PiyoPiyoPlayer(coreFile, coreMixer);
    public final PlayerControls playerControls = new PlayerControls(corePlayer);
    private final WindowCreatingUIElementConsumer worldCreator = new WindowCreatingUIElementConsumer();

    // Keeps the audio & editor threads from causing each other trouble in a simple way.
    // Note that this ABSOLUTELY HAS TO be on FairSync.
    // As it turns out, the audio thread tends to oversleep otherwise,
    //  and things to downhill from there.
    private final ReentrantLock coreFileLock = new ReentrantLock(true);
    public LinkedList<PiyoPiyoFile> undoBuffer = new LinkedList<PiyoPiyoFile>();
    public LinkedList<PiyoPiyoFile> redoBuffer = new LinkedList<PiyoPiyoFile>();

    public String currentModification = null;

    private double recommendedDeltaTime() {
        if (!playerControls.playing)
            return 0.2;
        // This had a lock in it, but locking is causing issues.
        // This is read-only so it's not an issue.
        double beatTime = corePlayer.piyoFile.beatTime();
        double almostFinal = beatTime - corePlayer.withinBeatPosition;
        if (playerControls.reverse)
            almostFinal = corePlayer.withinBeatPosition;
        return playerControls.halfSpeed ? (almostFinal * 2) : almostFinal;
    }

    public void runFrame() {
        // Upper bound of 20FPS. No more.
        try {
            double wakeUpTime = recommendedDeltaTime();
            while (wakeUpTime > 0.05d)
                wakeUpTime /= 2.0d;
            double deltaTime = GaBIEn.timeDelta(false);
            double remainingDeltaTime = wakeUpTime - deltaTime;
            long ms = (long) (remainingDeltaTime * 1000);
            if (ms > 0)
                Thread.sleep(ms + 5);
            while (GaBIEn.timeDelta(false) < wakeUpTime)
                Thread.yield();
        } catch (InterruptedException e) {
            e.printStackTrace();
        }

        double deltaTime = GaBIEn.timeDelta(true);

        if (worldCreator.runningWindows().size() == 0)
            GaBIEn.ensureQuit();

        coreFileLock.lock();
        try {
            worldCreator.runTick(deltaTime);
        } catch (Exception e) {
            if (currentModification != null) {
                currentModification = null;
                if (undoBuffer.size() > 0)
                    coreFile.copyFrom(undoBuffer.removeLast());
            }
            String saveState = "Data has been lost.";
            try {
                OutputStream os = GaBIEn.getOutFile("PPJ-BACKUP.pmd");
                coreFile.saveFile(os);
                os.close();
                saveState = "Data saved to 'PPJ-BACKUP.pmd'.";
            } catch (Exception e3) {
            }
            try {
                switchError(e, new NotesheetGame(this));
            } catch (Exception e2) {
                switchError(e, new UILabel("Exception occurred during recovery. " + saveState, 16));
            }
        }
        coreFileLock.unlock();
    }

    public void resetInitLauncher() throws IOException {
        outerGame = new UIGame();
        worldCreator.accept(outerGame);
        NotesheetGame ng = new NotesheetGame(this);
        try {
            PiyoPiyoPlayingSample.ensureAllSamplesLoaded();
            switcher(ng);
        } catch (IOException ioe) {
            switchError(ioe, ng);
        }
        GaBIEn.getRawAudio().setRawAudioSource(new IRawAudioDriver.IRawAudioSource() {
            @Override
            public short[] pullData(int samples) {
                coreFileLock.lock();
                short[] allData = new short[samples * 2];
                int sPtr = 0;
                // With this it's set at 1/100th precision.
                // So it doesn't lose track of time on Buriki.
                int magic = 220;
                while (samples > magic) {
                    short[] t1 = coreMixer.pullData(magic);
                    System.arraycopy(t1, 0, allData, sPtr * 2, magic * 2);
                    samples -= magic;
                    sPtr += magic;
                    double dT = magic / 22050d;
                    playerControls.runTick(dT);
                }

                short[] t1 = coreMixer.pullData(samples);
                System.arraycopy(t1, 0, allData, sPtr * 2, samples * 2);
                double dT = samples / 22050d;
                playerControls.runTick(dT);
                coreFileLock.unlock();
                return allData;
            }
        });
    }

    public void switcher(UIElement returnGame) {
        outerGame.changeElement(returnGame);
    }

    public void switchError(Exception e, final UIElement notesheetGame) {
        String ef = "Exception occurred! Details follow...\n";
        StringWriter sw = new StringWriter();
        PrintWriter pw = new PrintWriter(sw);
        e.printStackTrace(pw);
        pw.flush();
        ef += sw.toString().replace("\r\n", "\n");
        switcher(new UISplitterLayout(new UILabel(ef, 16), new UITextButton("Continue", 16, new Runnable() {
            @Override
            public void run() {
                switcher(notesheetGame);
            }
        }), true, 1));
    }

    public void startModifications(String s) {
        redoBuffer.clear();
        undoBuffer.add(new PiyoPiyoFile(coreFile));
    }

    public void endModifications() {
        if (currentModification != null)
            currentModification = null;
    }

    public void undoModifications() {
        if (undoBuffer.size() > 0) {
            redoBuffer.add(new PiyoPiyoFile(coreFile));
            coreFile.copyFrom(undoBuffer.removeLast());
        }
    }

    public void redoModifications() {
        if (redoBuffer.size() > 0) {
            undoBuffer.add(new PiyoPiyoFile(coreFile));
            coreFile.copyFrom(redoBuffer.removeLast());
        }
    }
}
